/********************************************************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *         with pair hidden Markov models, pair stochastic context-free
 *        grammars, and probabilistic evolutionary  models.
 *       
 * Version 2.0.0 (JUN 2003)
 *
 * Copyright (C) 2000-2003 Howard Hughes Medical Institute/Washington University School of Medicine
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ***********************************************************************************************************/

/* evolve.c
 *
 * ER, Fri Nov 22 10:53:20 CST 2002 [St. Louis]
 * 
 * Purpose:  Given:
 *                       - a set of joint probabilities P(a,b| t) at a given evolutionary time t,
 *
 *                       - set of conditional probabilities at time zero Q_0
 *         
 *           Calculate:
 *                       - P(a,b| r*t) using a Markov chain model for evolution
 *  
 *
 * Method:   (1) conditionals: P(a,b | t) --> Q(a | b,t) 
 *            
 *           (2) Q(a | b, r*t) = Q_0 * exp{r * log[Q_0^{-1} * Q}
 *
 *           (3) Q(a | b, r*t) --> P(a,b | r*t)
 *
 *
 *
 * ConditionalsEvolved()  creates:    Q(a | b, r*t)   from    P(a,b | t)
 *
 * Joint2Joint()          converts:   P(a | b, t)     to      P(a,b | r*t)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

static void    adjust_prob(double *psingle, int size);
static void    check_three_probs (double *T, double *T_zero, double *T_infty, int dim);
static void    isQconsistent(FILE *ofp, double *Q, int L, int pedantic, int verbose);
static void    islogQconsistent(FILE *ofp, double *Q_0, double *K, int Lx, int Ly, int Lz, int verbose);
static void    islogQconsistent_2(FILE *ofp, double *Q_0, double *A, double *K, int Lx, int Ly, int Lz, int verbose);




int
Check_Accuracy(double *vec, int L)
{
  int    i;
  int    issig;
  double m;

  issig = FALSE;
  
  for (i = 0; i < L; i ++) 
    {
      m = (vec[i] > 0)? vec[i]:-vec[i];
      if (m > (1.-accuracy)/(1.*L)) {
	issig = TRUE;
	break;
      }
    }
  
  return issig;
}

/* Function: Condi_From_Joint()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  Given       a LL matrix M(i,j) of joint probabilities, 
 *           calculate the LL matrix Q(i,j) of conditional probabilities.
 *
 *           Q[i*L+j] = Q(j|i) = P(i,j|t_o) / \sum_k  P(i,k|t_o) 
 *
 *           where P(i,j|t_o) are calculated according to 
 *           using a given pammodel t_o.
 *
 * Args:     P  - LxL joint prob matrix (prealloc)
 *
 * Returns:  Q(LxL), conditional probabilities.
 *           Q is alocated here, freed by caller.
 */
double *
Condi_From_Joint(FILE *ofp, double *P, int L, int verbose)
{
  double *Q;
  double *pm;
  double  sum;
  int     i, j;
  
  if (verbose) {
    fprintf(ofp, "P(i,j, t_o) probabilities\n");
    PrintProbs(ofp, P, L);
  }
  
  /* allocate Q[L*L]
   */
  Q  = (double *) MallocOrDie (sizeof(double) * L * L);
  pm = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++)
    pm[i] = 0.0;
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      pm[i] += P[i*L+j];
  
  for (i = 0; i < L; i++) {
    sum = 0.;
    for (j = 0; j < L; j++) {
      Q[i*L+j] = P[i*L+j] / pm[i];
      sum += Q[i*L+j];
    }
    /* normalize */
    for (j = 0; j < L; j++) 
      Q[i*L+j] /= sum;
  }
 
  if (verbose) { 
    fprintf(ofp, "Q(j|i, t_o) probabilities\n");
    PrintProbs(ofp, Q, L);
  }
  
  /* consistency check, rows should add up to one
   */
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  free (pm);

  return Q;
}

 
/* Function: ConditionalsEvolved()
 * Date:     ER, Wed Apr 12 18:11:23 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) using a Markov chain model for evolution
 *  
 * Method:   Q(a | b, r*t) = Q_0 * exp{r * log[Q_0^{-1} * Q}
 *
 *
 *
 *
 * Args:              
 *
 * Returns: Qr(LxL). 
 *          Qr is allocated here, freed by caller.
 *           
 */
void
ConditionalsEvolved(FILE *ofp, double *Q, double *Q_0, int L, double tfactor, int pedantic, int verbose)
{
  double  *Q_0_inv;
  double  *Qr;
  double  *K;
  double  *S;
  int      i;

  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  if (verbose) {
    fprintf(ofp, "Q(j|i, t_*)  probabilities\n");
    PrintProbs(ofp, Q, L);
  }
  /* calculate Q_0_inv
   */
  Q_0_inv = Cal_M_Inv(ofp, Q_0, L, verbose);
  
  /* construct K = log [Q_0_inv * Q]
   */
  Comp_M_N_Prod(ofp, Q_0_inv, Q, L, verbose); /*  multiply Q_0_inv*Q, dump it in Q */
  
  S = Cal_Id(L);
  Comp_M_N_Sum(ofp, Q, S, L, L, FALSE, verbose);   /* S = Q - Id  */
 
  if (verbose) 
    isQconsistent(ofp, S, L, pedantic, verbose);  /* check that Q is consistent with a markovian stationary 
						   * model of evolution. 
						   * In plain words it checks whether logQ is going to exist.
						   *
						   * Basically, it checks whether  all eigenvalues of Q - I
						   * are between -1 and 1. 
						   *
						   * remember log (1+x) = x - x^2/2 + ...   converges for -1 < x <= 1
						   *
						   */
  K = Cal_M_Log(ofp, Q, L, verbose);
 
  if (pedantic)
    islogQconsistent(ofp, Q_0, K, L, L, L, verbose); /*  check that K is consistent with a 
						      *  markovian stationary model of evolution. 
						      *
						      *  Basically, it checks whether all entries of  
						      *  Q_0 + epsilon*Q_0*K are none negatives.
						      */
  
  /* calculate Qr = Q_0 * exp{tfactor*K}
   */
  Qr = Cal_M_Exp(ofp, K, L, tfactor, verbose);
  
  Comp_M_N_Prod(ofp, Q_0, Qr, L, verbose); /* multiply Q_0*Qr, dump it in Qr */
  
  for (i = 0; i < L; i++)
    adjust_prob(Qr+i*L, L);
  
  if (verbose) {
    fprintf(ofp, "Q(j|i, r*t_o=%f)  probabilities\n", tfactor);
    PrintProbs(ofp, Qr, L);
  }
  
  CopyMatrix(Q, Qr, L, L);

  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  free(S);
  free(K);
  free(Q_0_inv);
  free(Qr);

}


/* Function: ConditionalsEvolved_2()
 * Date:     ER, Mon May  8 10:33:18 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) 
 *           this is a more general model than the previous one:
 *  
 * Method:   Q(a | b, r*t) = Q_0 +  exp{ r * log[I + R^{-1}(Q-Q_0)] }
 *
 *           for once, the property Q(t+s) = Q(t)*Q(s) is not satisfied any more.
 *
 * Args:              
 *
 * Returns: Qr(LxL).  *          Qr is allocated here, freed by caller.
 *           
 */
void
ConditionalsEvolved_2(FILE *ofp, double *Q, double *Q_0, double *R, int L, double tfactor, 
		      int pedantic, int verbose)
{
  double  *R_inv;
  double  *Qr;
  double  *K;
  double  *Id;
  int      i;

  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  Id  = Cal_Id(L);
  
  /* calculate R_inv 
   */
  R_inv = Cal_M_Inv(ofp, R, L, verbose);
  
  /* calculate I + R_inv*(Q-Q_0)  dump it in Q 
   */
  Comp_M_N_Sum(ofp, Q, Q_0, L, L, FALSE, verbose);    /*  add:      Q - Q_0,   dump it in Q */
  CopyMatrix(Q, Q_0, L, L);
  Comp_M_N_Prod(ofp, R_inv, Q, L, verbose);           /*  multiply: R_inv * Q, dump it in Q */

  if (pedantic)
    isQconsistent(ofp, Q, L, pedantic, verbose);  /* check that Q is consistent with a markovian stationary 
						   * modelof evolution. 
						   * In plain words it checks whether log(I+Q) is going to exist.
						   *
						   * Basically, it checks whether  all eigenvalues of Q
						   * are between -1 and 1. 
						   *
						   * remember log (1+x) = x - x^2/2 + ...   converges for -1 < x <= 1
						   *
						   */

  Comp_M_N_Sum(ofp, Id, Q, L, L, TRUE, verbose);   /*  add:      Id + Q,    dump it in Q */

  /* construct K = log [I + R_inv*(Q-Q_0)]   */
  K = Cal_M_Log(ofp, Q, L, FALSE);
  
  if (pedantic)
    islogQconsistent_2(ofp, Q_0, R, K, L,  L,  L, verbose); /*  check that K is consistent with a 
							     *  markovian stationary model of evolution. 
							     *
							     *  Basically, it checks whether all entries of  
							     *  Q_0 + epsilon*R*K are none negatives.
							     */
  
  /* calculate Qr = Q_0 + R * [ exp{tfactor*K} - Id ]
   */
  Qr = Cal_M_Exp(ofp, K, L, tfactor, FALSE);
  
  Comp_M_N_Sum(ofp, Qr, Id, L, L, FALSE, FALSE);     /* add:      Qr-Id,  dump it first in Id, then in Qr */
  CopyMatrix(Qr, Id, L, L);
  Comp_M_N_Prod(ofp, R, Qr, L, FALSE);               /* multiply: R*Qr,   dump it in Qr */
  Comp_M_N_Sum(ofp, Q_0, Qr, L, L, TRUE, FALSE);     /* add:      Q_0+Qr, dump it in Qr */
  
  
  if (verbose) {
    fprintf(ofp, "Q(j|i, r*t_o)  probabilities\n"); 
    PrintProbs(ofp, Qr, L);
  }
  
  CopyMatrix(Q, Qr, L, L);

  /* consistency check, rows should add up to one
   */	
  for (i = 0; i < L; i++)
    CheckSingleProb(Q+i*L, L);
  
  free(K);
  free(Id);
  free(R_inv);
  free(Qr);

}

/* Function: Joint_From_Condi()
 * Date:     ER, Mon Jun 12 14:41:47 CDT 2000 [St. Louis]
 *
 * Purpose:  Given       a LxL matrix Q(i,j) of conditional probabilities, 
 *           calculate the LxL matrix P(i,j) of joint       probabilities.
 *
 *           P(i,j) = Q(j|i) * pm[j] 
 *
 *           where P(i,j|t_o) are calculated according to 
 *           using a given pammodel t_o.
 *
 * Args:     Q  - LxL conditional prob matrix (prealloc)
 *           pm - L   single emision probabilities (prealloc)
 *
 * Returns:  P(LxL), conditional probabilities.
 *           P is alocated here, freed by caller.
 */
double *
Joint_From_Condi(FILE *ofp, double *Q, double *pm, int L, int verbose)
{
  double *P;
  int     i, j;
  
  if (verbose) {
    fprintf(ofp, "Q(i|j,t) probabilities\n");
    PrintProbs(ofp, Q, L);
  }
  
  /* allocate P[L*L]
   */
  P  = (double *) MallocOrDie (sizeof(double) * L * L);
 
  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) 
      P[i*L+j] = Q[i*L+j] * pm[i]; 
  
  if (verbose) { 
    fprintf(ofp, "P(i,j,t) probabilities\n");
    PrintProbs(ofp, P, L);
  }
  
  /* consistency check
   */
  CheckSingleProb(P, L*L);
  
  return P;
}

 
/* Function: Joint2Joint()
 * Date:     ER, Wed Apr 12 18:11:23 CDT 2000 [St. Louis]
 *
 * Purpose:  Given a set of joint probabilities P(a,b| t_o) 
 *           at a given evolutionary time t_o, 
 *           and set of conditional probabilities at time zer0 P^o
 *         
 *           calculate P(a,b| tfactor*t_o) using a Markov chain model for evolution
 *  
 * Method:   (1) calculate conditionals: P(a,b | t_o) --> Q(a | b,t_o) 
 *            
 *           (2) calculate marginals: P(a,b | t_o) --> pmar(a | b,t_o)  
 *            
 *           (3) evolve conditionals Q(a | b, r*t_o) = Q_0 * exp{ r * log[Q_0^{-1} * Q] }
 *
 *           (4) construct joints using same marginals Q(a | b, r*t_o), pmar(a | b,t_o)  --> P(a,b | r*t_o)
 *
 * Args:              
 *
 * Returns:  (void)
 *           
 */
void
Joint2Joint(double *p_joint, double *q_nought, int L, double tfactor, int pedantic, int verbose)
{
  double *pm;
  double *p_cond;
  double  p_diff;
  double  sum;
  int     i, j;
  
  /* allocate pm[L]
   */
  pm = (double *) MallocOrDie (sizeof(double) * L);

  for (i = 0; i < L; i++)
    pm[i] = 0.;
  
  sum = 0.;
  for (i = 0; i < L; i++) {
    for (j = 0; j < L; j++) 
      pm[i] += p_joint[i*L+j];
    sum += pm[i];
  }

  /* normalize */
  for (i = 0; i < L; i++) pm[i] /= sum;

  if (verbose) {
    printf("marginalized probabilities\n");
    for (i = 0; i < L; i++) 
      printf("%f ", pm[i]);
    printf("\n");
  }
  
  /* calculate the matrix of conditional probabilities (Q) from joint probabilities (P)
   */
  p_cond = Condi_From_Joint(stdout, p_joint, L, verbose);
  
  /* evolve conditionals
   */
  ConditionalsEvolved(stdout, p_cond, q_nought, L, tfactor, pedantic, verbose);
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      p_diff  += (p_joint[i*L+j] - p_cond[i*L+j]*pm[i]) * (p_joint[i*L+j] - p_cond[i*L+j]*pm[i]);

  if (verbose) printf("Joint Prob's [dim=%d]  difference from t=1 squared = %f\n", L, p_diff);
 
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      p_joint[i*L+j] = p_cond[i*L+j] * pm[i];

  CheckSingleProb(p_joint, L*L);
  
  free(pm);
  free(p_cond);
}

/* Function: Joint2JointGap()
 * Date:     ER, Tue Jan 14 11:24:14 CST 2003 [St. Louis]
 *
 * Purpose:  Given a set of joint probabilities P(a,b| t_*) 
 *           at a given evolutionary time t_*, 
 *           the set of conditional probabilities at time zero P^o,
 *           and the set of single probs at the time of interests pv
 *         
 *           calculate P(a,b| tfactor*t_o) using a Markov chain model for evolution.
 *           For probabilities that include gaps.
 *  
 * Method:   (1) calculate conditionals: P(a,b | t_*) --> Q(a | b,t_*) 
 *            
 *           (4) evolve conditionals Q(a | b, r*t_*) = Q_0 * exp{ r * log[Q_0^{-1} * Q] }
 *
 *           (6) construct joints with new probs Q(a | b, r*t_*), pv(a | b, r*t_*)  --> P(a,b | r*t_*)
 *
 * Args:              
 *
 * Returns:  (void)
 *           
 */
void
Joint2JointGap(double *p_joint, double *q_nought, double *pv, int L, double tfactor, int pedantic, int verbose)
{
  double *p_cond;
  double  sum;
  double  alpha;
  int     i, j;
  
   
  /* calculate the matrix of conditional probabilities (Q) from joint probabilities (P)
   */
  p_cond = Condi_From_Joint(stdout, p_joint, L, verbose);
  
  /* evolve conditionals
   */
  ConditionalsEvolved(stdout, p_cond, q_nought, L, tfactor, pedantic, verbose);
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      p_joint[i*L+j] = p_cond[i*L+j] * pv[i];

  CheckSingleProb(p_joint, L*L);

  if (verbose) {
    fprintf(stdout, "P_COND(i,j| r*t_o=%f)  Cond probabilities\n", tfactor);
    PrintProbs(stdout, p_cond, L);
    fprintf(stdout, "Psingle(i| r*t_o=%f) single nt probabilities\n", tfactor);
    PrintVectorProbs(stdout, pv, L);
    fprintf(stdout, "P_JOINT(i,j| r*t_o=%f)  Joint probabilities\n", tfactor);
    PrintProbs(stdout, p_joint, L);
 }
    
  free(p_cond);
}



/* Function: TimeIdCorrelation()
 *
 * Date:     ER, Thu Nov 14 14:16:12 CST 2002 [St. Louis]
 *
 *
 *     id = 100 - div.a * time ^ div.b ==>  log (time) = 1/div.b log[(100-id)/div.a]
 *
 */
double 
TimeIdCorrelation(struct divergence_s div, double id)
{
  double time;
  double id_cutoff = 0.0;

  if (div.a <=   0.) { time = 1.0; return time; }
  if (id    == 100.) { time = 0.0; return time; }

  if (id >= id_cutoff)
    time = LOG2( (100.0-id) / div.a);
  else
    time = LOG2( (100.0-id_cutoff) / div.a);

  if (div.b != 0.0) 
    time /= div.b;
  else  
    Die ("TimeIdCorrelation() error. Division by zero\n");

  time = EXP2(time);

  return time;
}

/* Function: TransitionsEvolvedLinear()
 * Date:     ER, Fri Jul 26 15:20:09 CDT 2002 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) using a Markov chain model for evolution
 *  
 * Method:   
 *
 *                    q^i(t) = q^i_0 + r^i [ exp{tA^i} - I ]  ---   A(nxn), q(1xn), q_0(1xn), r(1xn)
 *
 *
 *              | q^1 |                           | q^1_o |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *        Q^* = | q^i |                     Q_0 = | q^i_o |               R^i = |  0  |          A^i = 1/t* log [ | 1 | (q^i - q^I_o) + I ]
 *              |  .  |                           |   .   |                     | r^i |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              | q^n |                           | q^n_o |                     |  0  |                           | 1 |
 *
 *
 *
 *     THEN,      Q(t)  =  Q_0  + Sum_i R^i [ exp(tA^i) - I ]
 *
 * Args:              
 *
 * Returns: Qr(LxL). 
 *          Qr is allocated here, freed by caller.
 *           
 */
void
TransitionsEvolvedLinear(FILE *ofp, double *q, double *q_zero, double *r, double *r_diag, double *r_fix, 
			 int L, double tfactor, int pedantic, int verbose)
{
  double  *Id;        /* LxL array */
  double  *A;         /* LxL array */
  double  *K;         /* LxL array */
  double  *Qr;        /* 1xL array */
  int      i, j;

  Id = Cal_Id(L);

  /* Allocate 
   */
  A = (double *) MallocOrDie (sizeof(double) * L * L);

  /* 
   *
   *                       | 1 |           | K1                |
   *                       | 1 |           |    K2             |
   *                       | 1 |           |       K3          |
   *    Calculate  A = I + | 1 | r_fix +   |          .        | 
   *                       | 1 |           |            .      | 
   *                       | 1 |           |              .    | 
   *                       | 1 |           |                Kn | 
   *
   *
   *          with Ki = r_diag^i =  ( q^i - q^i_zero - beta * r^i_fix) / r^i
   *
   */

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) {
      A[i*L+j] = r_fix[j];  

      if (i==j) A[i*L+j] += r_diag[i];
    }
  
  if (pedantic)
    isQconsistent(ofp, A, L, pedantic, verbose);  /* check that A is consistent with a markovian stationary 
						   * model of evolution. 
						   * In plain words it checks whether log(I+A) is going to exist.
						   *
						   * Basically, it checks whether  all eigenvalues of A
						   * are between -1 and 1. 
						   *
						   * remember log (1+x) = x - x^2/2 + ...   converges for -1 < x <= 1
						   *
						   */

  /* Now add the identity to complete the 
   * definition of A
   */
  for (i = 0; i < L; i++) A[i*L+i] += 1.0;

  if (FALSE) {
    fprintf(ofp, "\n EXP[ t^* A ] \n");

    for (i = 0; i < L; i++) {
      for (j = 0; j < L; j++) 
        fprintf(ofp, "%.4f ", A[i*L+j]);
      fprintf(ofp, "\n");
    }
    fprintf(ofp, "\n\n");
  }

  /* construct K = log [A]
   */
  K = Cal_M_Log(ofp, A, L, verbose);
  
  if (pedantic)
    islogQconsistent_2(ofp, q_zero, r, K, 1, L, L, verbose); /*  check that K is consistent with a 
							      *  markovian stationary model of evolution. 
							      *
							      *  Basically, it checks whether all entries of  
							      *  q_zero + epsilon*q_zero*K are none negatives.
							      */
  
  /* calculate Qr = Q_0 + R * [ exp{tfactor*K} - Id]
   */    
  Comp_M_Exp(ofp, K, L, tfactor, verbose);                /* exp:      exp(t*K),  dump it in K      */
  Comp_M_N_Sum(ofp, K, Id, L, L, FALSE, verbose);             /* add:      K-Id,      dump it in K      */
  CopyMatrix(K, Id, L, L);
  Qr = Cal_M_N_Prod(ofp, r, K, 1, L, L, verbose);             /* multiply: r*K,       dump it in Qr     */
  Comp_M_N_Sum(ofp, q_zero, Qr, 1, L, TRUE, verbose);         /* add:      q_zero+Qr, dump it in Qr     */
  
  
  if (TRUE) {
    fprintf(ofp, "\nT(i->j | t)  Transition probabilities at time %.3f\n", tfactor);
    for (i = 0; i < L; i++) 
        fprintf(ofp, "%.4f ", Qr[i]);
    fprintf(ofp, "\n----------------------------------------------------------\n");
  }
  
  CopyMatrix(q, Qr, 1, L);
  /* consistency check, rows should add up to one
   */	
  CheckSingleProb(q, L);
  
  free(K);
  free(A);
  free(Id);
  free(Qr);

}

/* Function: TransitionsEvolved()
 * Date:     ER, Tue Aug 13 16:17:27 CDT 2002 [St. Louis]
 *
 * Purpose:  Given a set of conditionals probabilities Q(a|b, t) 
 *           at a given evolutionary time t, 
 *           and set of conditional probabilities at time zero Q_0
 *         
 *           calculate Q(a,b| r*t) using a Markov chain model for evolution
 *  
 * Method:   
 *
 *                    q^i(t) = q^i_0 + r^i [ exp{tA^i} - I ]  ---   A(nxn), q(1xn), q_0(1xn), r(1xn)
 *
 *
 *              | q^1 |                           | q^1_o |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *        Q^* = | q^i |                     Q_0 = | q^i_o |               R^i = |  0  |          A^i = 1/t* log [ | 1 | r^fix + Diag(e^(-K^1)) ]
 *              |  .  |                           |   .   |                     | r^i |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              |  .  |                           |   .   |                     |  0  |                           | 1 |
 *              | q^n |                           | q^n_o |                     |  0  |                           | 1 |
 *
 *
 *
 *     THEN,      Q(t)  =  Q_0  + Sum_i R^i [ exp(tA^i) - I ]
 *
 * Args:              
 *
 * Returns: Qr(LxL). 
 *          Qr is allocated here, freed by caller.
 *           
 */
void
TransitionsEvolved(FILE *ofp, double *q, double *q_zero, double *q_infty, double *r_diag, 
		   int L, double tfactor, int pedantic, int verbose)
{
  double  *Id;        /* LxL array */
  double  *A;         /* LxL array */
  double  *K;         /* LxL array */
  double  *Q;         /* 1xL array */
  double  *r;         /* 1xL array */
  double   w = 0.0;
  int      isdiagonal = 1;
  int      i, j;

  Id = Cal_Id(L);

  if (verbose) {
    fprintf(ofp, "\nT(i->j | t)  Transition probabilities at time zero\n");
    for (i = 0; i < L; i++) 
      fprintf(ofp, "%.4f ", q_zero[i]);
    fprintf(ofp, "\nT(i->j | t)  Transition probabilities at time star\n");
    for (i = 0; i < L; i++) 
      fprintf(ofp, "%.4f ", q[i]);
    fprintf(ofp, "\nT(i->j | t)  Transition probabilities at time infty\n");
    for (i = 0; i < L; i++) 
      fprintf(ofp, "%.4f ", q_infty[i]);
    printf("\n");
  }
  
  /* Allocate 
   */
  A = (double *) MallocOrDie (sizeof(double) * L * L);

  /* 
   *
   *                      | e^(-K^1)                            |
   *                      |         e^(-K2)                     |
   *                      |                e^(-K3)              |
   *    Calculate  A =    |                       .             | 
   *                      |                          .          | 
   *                      |                             .       | 
   *                      |                             e^(-Kn) | 
   *
   *
   *          with   e^(-K^i) = 1 + r_diag^i =  1 + ( q^i - q^i_zero) / r^i
   *
   */

  for (i = 0; i < L; i++) A[i*L+i] = 1.0 + r_diag[i];

  if (verbose) {
    fprintf(ofp, "\n EXP[ t^* A ] \n");

    for (i = 0; i < L; i++) {
      for (j = 0; j < L; j++) 
        fprintf(ofp, "%.4f ", A[i*L+j]);
      fprintf(ofp, "\n");
    }
    fprintf(ofp, "\n\n");
  }


  /* Calculate exp (t A), 
   *
   *   If A is diagonal it is very simple:  exp(tA) =  exp[ t^* A ] * t
   *
   *   Otherwise have to calculate the log and then exponentiate.
   *
   */
  if (isdiagonal) { 
    K = (double *) MallocOrDie (sizeof(double) * L * L);
    for (i = 0; i < L; i++) 
      for (j = 0; j < L; j++) 
	if (i==j) {
	  if (tfactor > 0.0) K[i*L+j] = exp(log(A[i*L+j]) * tfactor); 
	  else               K[i*L+j] = 1.0;
	}
	else
	  K[i*L+j] = 0.0;
  } 
  
  else {
    /* construct K = log [A]
     */
    K = Cal_M_Log(ofp, A, L, verbose);
    
    if (pedantic)
      islogQconsistent_2(ofp, q_zero, r, K, 1, L, L, verbose); /*  check that K is consistent with a 
								*  markovian stationary model of evolution. 
								*
								*  Basically, it checks whether all entries of  
							        *  q_zero + epsilon*q_zero*K are none negatives.
							        */
    
    /* exp{tfactor*K} 
     */    
    Comp_M_Exp(ofp, K, L, tfactor, verbose);                /* exp:      exp(t*K),  dump it in K      */
  }
  

  /* Calculate w(t) = u^T exp(tA) r
   */
  r = Cal_M_N_Sum(ofp, q_zero, q_infty, 1, L, FALSE, verbose);         /* add:      r = q_0 - q_infty,           */  

  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++)  
      w += K[i*L+j] * r[j];
 
  /* Calculate 
   *
   *           q(t) = q_0 + [ exp(tA) - I ] r
   *
   */
  Comp_M_N_Sum(ofp, K, Id, L, L, FALSE, verbose);                   /* add:      K-Id,      dump it in K      */
  CopyMatrix (K, Id, L, L);
  Q = Cal_M_N_Prod(ofp, r, K, 1, L, L, verbose);                    /* multiply: r*K,       dump it in q      */
  Comp_M_N_Sum(ofp, q_zero, Q, 1, L, TRUE, verbose);                /* add:      q_zero+q,  dump it in q      */
  
  
  /* Normalize the probabilities
   */
  for (i = 0; i < L; i++) 
    Q[i] /= (1. + w);

  if (verbose) {
    fprintf(ofp, "\nT(i->j | t)  Transition probabilities at time %.3f\n", tfactor);
    for (i = 0; i < L; i++) 
      fprintf(ofp, "%.4f ", Q[i]);
    fprintf(ofp, "\n----------------------------------------------------------\n");
  }
  
  CopyMatrix(q, Q, 1, L);

  /* consistency check, rows should add up to one
   */	
  CheckSingleProb(q, L);
  
  free(r);
  free(K);
  free(A);
  free(Id);
  free(Q);

 }


/* Function: TransitionsDirectoryCosines()
 * Date:     ER, Tue Aug 13 17:01:49 CDT 2002 [St. Louis]
 *
 * Purpose:  Fill R_diag ant T_rate of the transfer matrix for the OTH model
 *
 * Args:     T
 *           T_zero
 *
 *
 * Returns:  void. Calculates R_diag and T_rate. Both allocated here, freed by caller.
 */
void
TransitionsDirectoryCosines(double *q, double *q_zero, double *q_rate, double *R_diag, int dim)
{
  double *q_infty;
  double  sum = 0.0;
  double  angle;
  double  shift;
  int     num = 0;
  int     col;

  /* Allocate q^infty */
  q_infty = (double  *) MallocOrDie (sizeof(double) * dim);

  /* fill R_diag.  
   * 
   *  R_diag_i = cos(theta)
   * 
   *  cos (theta) = (q_i - q^0_i) / (q^0_i - q^infty_i) 
   * 
   * 
   *  -1 < cos(theta) <= 0 ===> pi/2 <= theta < pi --- 
   * 
   *  also because 0 <= q_0 - r <= 1  ===>  cos(theta) <= (q^* / q^0) - 1
   * 
   * 
   */
  angle = 3.0*PI/4.0;
  shift = 0.05;

  for (col = 0; col < dim; col++) {
    if (q[col] - q_zero[col] < MARGIN && q[col] - q_zero[col] > -MARGIN) { 
      R_diag[col] = 0.0;           
      sum += q_zero[col]; 
      num ++; 
    }
    else {
      if (cos(angle) <= q[col]/q_zero[col] - 1.0) R_diag[col] = cos(angle);
      else                                        R_diag[col] = q[col]/q_zero[col] - 1.0;
    }
  }
  
  /* fill T_rate.  
   * 
   * q^rate = q^infty - q^0
   * 
   * q^rate = (q_i - q^0_i) / cos (theta)
   * 
   *  If  cos (theta) = 0, q^rate is free with the condition    0 <= q^0 - q^rate <= 1
   * 
   */
  sum /= num;
  for (col = 0; col < dim; col++) {
    if (q[col] - q_zero[col] < MARGIN && q[col] - q_zero[col] > -MARGIN) 
      q_rate[col] = q_zero[col] - sum; 
    else {                                                                
      if (cos(angle) <= q[col]/q_zero[col] - 1.0) q_rate[col] = (q[col] - q_zero[col]) / R_diag[col];
      else                                        q_rate[col] = q_zero[col];
    }

    q_infty[col] = q_zero[col] - q_rate[col];
  }
 
  if (FALSE) {
    printf("RA_diag - OTH transfer matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%f ", R_diag[col]);
    
    printf("\n");
    
    printf("TA_rate - OTH transfer matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%f ", q_rate[col]);
    
    printf("\n");
    
    
    printf("TA_infty - OTH transfer matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%f ", q_infty[col]);
    
    printf("\n");
  }

  /* paranoia - check that q^infty = q^0 + q_rate add up to one
   */
  CheckSingleProb(q_infty, dim);
    
  free(q_infty);
}

/* Function: TransitionsExpVector()
 * Date:     ER, Wed Aug 14 17:48:05 CDT 2002 [St. Louis]
 *
 * Purpose:  Fill R_diag vector and r vector
 *                 
 *               from the expressions:        r_i = q^0_i - q^infty_i
 *
 *                                       R^diag_i = ( q_i - q^0_i ) / r_i
 *                 
 * Args:     
 *
 * Returns:  void
 */
double *
TransitionsExpVector(double *T, double *T_zero, double *T_infty, int dim)
{
  double  *R_exp;
  double   sum   = 0.;
  int      col;
 
  /* check the three input probability vectors are in good order
   *
   * to have things well behaving one of these two conditions have to occur
   *
   *
   *            T^0_i < T^*_i < T^infty_i    or     T^infty_i < T^*_i < T^0_i  
   *
   */
  /*check_three_probs (T, T_zero, T_infty, dim);*/

  /* allocate R_diag
   */
  R_exp = (double *) MallocOrDie (sizeof(double) * dim);

  /* initialize
   */
  for (col = 0; col < dim; col++)
    R_exp[col] = 0.0;
  
  /* Calculate 
   *
   *    
   *
   */
  for (col = 0; col < dim; col++)  
    if (T_zero[col] - T_infty[col] < MARGIN && T_zero[col] - T_infty[col] > -MARGIN) 
      R_exp[col] = 0.0;     
    else
      R_exp[col] = (T[col] - T_zero[col]) / (T_zero[col] - T_infty[col]); 
  
  if (FALSE) {
    printf("R^exp - matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%.4f ", R_exp[col]);
    printf("\n");
  }
  
  /* paranoia: 
   *
   *           Define the scalar product K.R = sum_k K[k] * R[k]
   *        
   *           IS R^diag.R = 0 ????
   *
   *           Vectors R and R^diag being orthogonals  guarantees that q(t) add up to one.
   *
   */
  for (col = 0; col < dim; col++)
     sum += (T_zero[col] - T_infty[col]) * R_exp[col];
  if (sum < -MARGIN && sum > MARGIN) Die ("Bad Diag Matrix, sum_col K[col] R[col] = %f", sum);
    
  return R_exp;
}

/* Function: TransitionsDiagVector()
 * Date:     ER, Fri Aug  9 15:43:08 CDT 2002 [St. Louis]
 *
 * Purpose:  Fill R_diag matrix 
 *                 
 *               from the expression: R^diag_i * r_i = q_i - q^0_i - (sum_l r_l) R^fix_i  
 *                 
 *                 
 *           R^diag_k = (q_k - q^0_k - R^fix_k ) / r_k
 *
 * Args:     R_diag -- transfer matrix for the trnasitions  of the model
 *
 * Returns:  R_diag is allocated and filled here. R_diag freed by caller.
 */
double *
TransitionsDiagVector(double *T, double *T_zero, double *T_rate, double *R_fix, int dim)
{
  double *R_diag;
  double  beta  = 0.;
  double  sum   = 0.;
  int     col;

  /* allocate R_diag
   */
  R_diag = (double *) MallocOrDie (sizeof(double) * dim);

  /* initialize
   */
  for (col = 0; col < dim; col++)
    R_diag[col] = 0.0;
  
  /* Calculate beta = sum_k R[k]
   *
   */
  for (col = 0; col < dim; col++) beta += T_rate[col];
  if (beta < -MARGIN || beta > MARGIN) Die ("Bad Rate Matrix in TransitionsDiagVector(), sum_col R[col] = %f", beta);
  
  /* Calculate R^diag
   *
   *                       R^diag_k = (q_k - q^0_k - R^fix_k ) / r_k
   *
   */
  for (col = 0; col < dim; col++) 
    R_diag[col] = (T[col] - T_zero[col] - R_fix[col]) / T_rate[col]; 
  
  if (FALSE) {
    printf("R^diag - matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%.4f ", R_diag[col]);
    printf("\n");
  }
  
  /* paranoia: 
   *
   *           Define the scalar product K.R = sum_k K[k] * R[k]
   *        
   *           IS R^diag.R = 0 ????
   *
   *           Vectors R and R^diag being orthogonals  guarantees that q(t) add up to one.
   *
   */
  for (col = 0; col < dim; col++)
     sum += T_rate[col] * R_diag[col];
  if (sum < -MARGIN && sum > MARGIN) Die ("Bad Diag Matrix, sum_col K[col] R[col] = %f", sum);
    
  return R_diag;
}


/* Function: TransitionsFixVector()
 * Date:     ER, Fri Aug  9 15:32:29 CDT 2002 [St. Louis]
 *
 * Purpose:  Fill R_fix matrix 
 *                 
 *               from the expression: R^diag_i * r_i = q_i - q^0_i - (sum_l r_l) R^fix_i  
 *                 
 *                 
 *           R^fix_k = (q_k - q^0_k) / beta - R^diag_k * r_k
 *
 *           where  beta = (sum_l r_l) 
 *
 * Args:     R_fix -- transfer matrix for the trnasitions  of the model
 *
 * Returns:  R_fix is allocated and filled here. R_fix freed by caller.
 */
double *
TransitionsFixVector(double *T, double *T_zero, double *T_rate, double *R_diag, int dim)
{
  double *R_fix;
  double  beta  = 0.;
  double  sum   = 0.;
  int     col;
  
  /* allocate R_fix
   */
  R_fix = (double *) MallocOrDie (sizeof(double) * dim);
  
  /* initialize
   */
  for (col = 0; col < dim; col++)
    R_fix[col] = 0.0;
  
  /* Calculate beta = sum_k R[k]
   *
   */
  for (col = 0; col < dim; col++) beta += T_rate[col];
  if (beta < -MARGIN && beta > MARGIN) Die ("Bad Rate Matrix in TransitionsFixVector(), sum_col R[col] = %f", beta);
  
  /* Calculate R^fix
   *
   *        R^fix_k = (q_k - q^0_k) / beta - R^diag_k * r_k
   *  
   */
  for (col = 0; col < dim; col++) 
    R_fix[col] = (T[col] - T_zero[col]) / beta - R_diag[col] * T_rate[col]; 
  
  if (FALSE) {
    printf("R^fix - matrix\n");
    for (col = 0; col < dim; col++) 
      printf("%.4f ", R_fix[col]);
    printf("\n");
  }
  
  /* paranoia: 
   *            elements of R^fix should add up to zero
   *
   */
  for (col = 0; col < dim; col++)
    sum += R_fix[col];
  if (sum < -MARGIN || sum > MARGIN) Die ("Bad Fix Matrix, sum_col R_fix[col] = %f", sum);
  
  return R_fix;
}


/* Function: TransitionsRateVector()
 * Date:     ER, Wed Aug  7 09:52:32 CDT 2002 [St. Louis]
 *
 * Purpose:  Fill R the rate matrix 
 *                 
 *               from the expression: R^diag_i * r_i = q_i - q^0_i - (sum_l r_l) R^fix_i  
 *                 
 *                 
 *           r_k = (q_k - q^0_k) / R^diag_k - beta * R^fix_k / R^diag_k
 *
 *           where  beta = (sum_l r_l) = [ sum_l (q_l - q^0_l) / R^diag_l ] / 1 + sum_l R^fix_k / R^diag_k
 *
 * Args:     R -- transfer matrix for the trnasitions  of the model
 *
 * Returns:  R is allocated and filled here. R freed by caller.
 */
double *
TransitionsRateVector(double *T, double *T_zero, double *R_fix, double *R_diag, int dim)
{
  double *R;
  double  beta  = 0.;
  double  sum_q = 0.;
  double  sum_i = 0.;
  double  sum   = 0.;
  int     col;

  /* allocate R 
   */
  R = (double *) MallocOrDie (sizeof(double) * dim);
  
  /* initialize
   */
  for (col = 0; col < dim; col++)
    R[col] = 0.0;
  
  /* Calculate beta = sum_k R[k]
   *
   */
  for (col = 0; col < dim; col++) 
    if (R_diag[col] > MARGIN || R_diag[col] < -MARGIN) {
      sum_q += (T[col] -  T_zero[col]) / R_diag[col]; 
      sum_i += R_fix[col]              / R_diag[col];
    }
  
  if (sum_q != 0.) {
    if (1.0 + sum_i < MARGIN && 1.0 + sum_i > -MARGIN) 
      Die ("There is an inconsistency in your Rate matrix. Revise your directory cosines. sum_q = %f   1+sum_i = %f", sum_q, 1.0+sum_i);
    else
      beta += sum_q / (1.0 + sum_i);
  }
  
  /* fill R 
   *           R_k = ( q_k - q^0_k - beta R^fix_k ) / R^diag_k
   *  
   */
  for (col = 0; col < dim; col++) 
    if (R_diag[col] > MARGIN || R_diag[col] < -MARGIN) 
      R[col] += ( T[col] -  T_zero[col] - beta * R_fix[col] ) / R_diag[col]; 
  
  if (FALSE) {
    printf("R - rate matrix -  beta = %f\n", beta);
    for (col = 0; col < dim; col++) 
      printf("%.4f ", R[col]);
    printf("\n");
  }
  
  /* paranoia: 
   *
   *           Define the scalar product K.R = sum_k K[k] * R[k]
   *        
   *           IS R^diag.R = 0 ????
   *
   *           Vectors R and R^diag being orthogonals  guarantees that q(t) add up to one.
   *
   */
  for (col = 0; col < dim; col++)
    sum += R[col] * R_diag[col];
  if (sum < -MARGIN || sum > MARGIN) Die ("Bad Rate Matrix, sum_col K[col] R[col] = %f", sum);
  
  return R;
}

/* Function: check_three_probs()
 * Date:     ER, Wed Aug 14 18:12:30 CDT 2002 [St. Louis]
 *
 * Purpose:  check the three input probability vectors are in good order
 *
 *             to have things well behaving one of these two conditions have to occur
 *
 *
 *              T^0_i < T^*_i < T^infty_i    or     T^infty_i < T^*_i < T^0_i  
 *
 */
void
check_three_probs (double *T, double *T_zero, double *T_infty, int dim)
{
  int i;

  /* check for one of the two conditions
   *
   *   q_0 < q_* < q_infty    or  q_0 > q_* > q_infty
   */
  for (i = 0; i < dim; i++) 
    if      (T_infty[i] <= T[i] && T[i] <= T_zero[i]) ;
    else if (T_infty[i] >= T[i] && T[i] >= T_zero[i]) ;
    else Die ("check your parameters for state %d. Time zero:(%f), star:(%f) and infty:(%f) \n", i, T_zero[i], T[i], T_infty[i]);
  
  
  /*
   * Also check that if two of the probs are equal, the third it is also equal.
   */
  for (i = 0; i < dim; i++) {
    if ( T_zero[i] - T[i]  < MARGIN*MARGIN && T_zero[i]  - T[i] > -MARGIN*MARGIN  &&
	 (T_infty[i] - T[i] > MARGIN*MARGIN || T_infty[i] - T[i] < -MARGIN*MARGIN)  )
      Die ("check your parameters for state %d. Time zero:(%f), star:(%f) and infty:(%f) \n", i, T_zero[i], T[i], T_infty[i]);
  }
}


/* Function: isQconsistent()
 * Date:     ER, Thu Apr 20 14:19:42 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 * exp {t K}
 *
 *
 *           IF: Q_0 + epsilon * Q_0*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:      Q_0 - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  1 if consistent, 0 otherwise.
 */
void
isQconsistent(FILE *ofp, double *Q, int L, int pedantic, int verbose)
{
  struct eigenvalue_s *eigen;
  double              *H;
  double               norm;
  int                  i;
  
  H = HessenbergForm(ofp, Q, L);
  Hessenberg2Eigenvalues(ofp, H, L, &eigen, pedantic, verbose);

  if (verbose) {
      fprintf(ofp, "Q matrix\n"); 
      PrintProbs(ofp, Q, L);

      fprintf(ofp, "Hessenberg form\n"); 
      PrintProbs(ofp, H, L);
      
      fprintf(ofp, "Eigenvalues for Q\n"); 
      for (i = 0; i < L; i++) 
	fprintf(ofp, "%.4f %.4f \n", eigen->real[i], eigen->imag[i]);
  }
  
  /* The condition is that ALL the eigenvalues of Q are of norm smaller
   * than one. 
   */
  for (i = 0; i < L; i++) {
    norm = sqrt(eigen->real[i]*eigen->real[i]+eigen->imag[i]*eigen->imag[i]);
      if (pedantic && norm > 1.) 
	Warn ("Q not consistent with a markovian stationary model of evolution.\n Norm of eigenvalue (%f, %f) is = %f", 
	      eigen->real[i], eigen->imag[i], norm);
}

  free(H);
  free(eigen->real);
  free(eigen->imag);
  free(eigen);
}
  
/* Function: islogQconsistent()
 * Date:     ER, Fri Apr 14 12:06:24 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 * exp {t K}
 *
 *
 *           IF: Q_0 + epsilon * Q_0*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:      Q_0 - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  void.
 */
void
islogQconsistent(FILE *ofp, double *Q_0, double *K, int Lx, int Ly, int Lz, int verbose)
{
  double *R;
  int     i, j;

  R = Cal_M_N_Prod(ofp, Q_0, K, Lx, Ly, Lz, FALSE); /*  multiply Q_0*K, dump it in R */

  for (i = 0; i < Lx; i++)
    for (j = 0; j < Lz; j++) 
      if ( (R[i*Lz+j] < 0.0 && Q_0[i*Lz+j] < 0.0)  )
	Warn ("Ah! Q_0log[Q_0^{-1}*Q] not consistent with a markovian stationary model of evolution (%d,%d: %f %f).", i, j, Q_0[i*Lz+j], R[i*Lz+j]);     
  
  if (verbose) { 
    fprintf(ofp, "RA Instant evolution probabilities\n");
    for (i = 0; i < Lx; i++) {
      for (j = 0; j < Lz; j++) {
	fprintf(ofp, "%.4f ", R[i*Lz+j]);
      }
      fprintf(ofp, "\n");
    }
  }
  
  free(R);
}

/* Function: islogQconsistent_2()
 * Date:     ER, Mon May  8 12:55:29 CDT 2000 [St. Louis]
 *
 * Purpose:  Q(t) ---a matrix of conditional probabilities---
 *           is consistent with being derived from a markovian stationary
 *           model of evolution from:
 *                    Q_0 - a matrix of conditional probabilities at time zero
 *                    A - a matrix of rate of evolution
 *                    K - a matrix of rate of evolution
 *
 *           as
 *                     Q(t) = Q_0 + A[ exp{t K} - Id ]
 *
 *
 *           IF: Q_0 + epsilon * A*K  can play the role of an "instant matrix of evolution"
 *
 *               that is, if all the entries are positive.
 *
 * Args:     Q_0  - LxL matrix of conditional prob matrix at time zero (prealloc)
 *           A    - LxL matrix  (prealloc)
 *           logQ - LxL log matrix of conditional prob matrix (prealloc)
 *
 * Returns:  void.
 */
void
islogQconsistent_2(FILE *ofp, double *Q_0, double *A, double *K, int Lx, int Ly, int Lz, int verbose)
{
  double *R;
  int     i, j;

  R = Cal_M_N_Prod(ofp, A, K, Lx, Ly, Lz, FALSE); /*  multiply A*K, dump it in R */

  for (i = 0; i < Lx; i++)
    for (j = 0; j < Lz; j++)
      if ( (R[i*Lz+j] < 0.0 && Q_0[i*Lz+j] < 0.0)  )
	Warn ("Ah! Q_0log[Q_0^{-1}*Q] not consistent with a markovian stationary model of evolution (%d,%d: %f %f).", i, j, Q_0[i*Lz+j], R[i*Lz+j]);     
  
  if (verbose) { 
    fprintf(ofp, "RA Instant evolution probabilities\n");
    for (i = 0; i < Lx; i++) {
      for (j = 0; j < Lz; j++) {
	fprintf(ofp, "%.4f ", R[i*Lz+j]);
      }
      fprintf(ofp, "\n");
    }
  }
  
  free(R);
}


void
adjust_prob(double *psingle, int size)
{
  int x;
  double sum = 0.0;

  for (x = 0; x < size; x++) {
    if (psingle[x] > 1.0+MARGIN) Die ("adjust_prob(): probabilities are getting too large here. P[%d] = %f", x, psingle[x]);
    if (psingle[x] < -MARGIN2)   Warn ("adjust_prob(): probabilities are getting too small here, P[%d] = %f", x, psingle[x]);

    if (psingle[x] < 0.0 && psingle[x] > - MARGIN2) 
      psingle[x] = 0.0;
    
   sum += psingle[x];
  }

  for (x = 0; x < size; x++) psingle[x] /= sum;
 
}

