/********************************************************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *         with pair hidden Markov models, pair stochastic context-free
 *        grammars, and probabilistic evolutionary  models.
 *       
 * Version 2.0.0 (JUN 2003)
 *
 * Copyright (C) 2000-2003 Howard Hughes Medical Institute/Washington University School of Medicine
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ***********************************************************************************************************/

/* matrix.c
 *
 * ER, Fri Nov 22 11:51:28 CST 2002 [St. Louis]
 * 
 * Purpose:  Includes all the functions for matrix manipulations
 *           used in constructing the evolutionary models.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif

/* Function: Cal_Id()
 * Date:     ER, Thu Apr 13 10:28:16 CDT 2000 [St. Louis]
 *
 * Purpose:  Creates a Id(LxL) matrix
 *         
 * Args:    L - dimension          
 *
 * Returns: Id(LxL) = \delta_{ij}
 *          Id is allocated here, freed by caller.
 *           
 */
double *
Cal_Id(int L)
{
  double *Id;
  int     i, j;

  Id = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      Id[i*L+j] = ((i == j) ? 1. : 0.);

  return Id;
}

void
Comp_Id(double *Id, int L)
{
  int i, j;

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      Id[i*L+j] = ((i == j) ? 1. : 0.);
}

/* Function: Cal_LUP_Decomp()
 * Date:     ER, Mon Apr 17 13:43:07 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculate a LUP decomposition, stored in LU
 *  
 *           (Implemented after reading chapter 31 of "introduction to algorithms"
 *            by Cormen et al.)
 *            Find matrices P, L(lower triangular), U(uppper triangular)
 *            such that P*M = L*U
 *
 *           if l_ij with i<j  are the elements of L  (l_ii = 1)
 *              u_ij with i>=j are the elements of U  (u_ii = m_ii)
 *
 *                 | l_ij for i  > j (we forget about the one's in the diagonal of L)
 *           LU  = |
 *                 | u_ij for i <= j
 *
 * Method: 
 *           Inspect by column k  
 *           Inspect rows from k on, until u_ik is a max value.
 *           Store the permutation in P.
 *           Permute rows k with i both in m and LU matrices.
 *
 *           u_ik =            m_ik - \sum_{j=1}^{j=i-1} [ l_ij * u_jk ]
 *
 *           l_ik = 1/u_kk * { m_ik - \sum_{j=1}^{j=k-1} [ l_ij * u_jk ]}
 *
 *
 *
 * Args:     M - LxL non-singular matrix.
 *
 * Returns:  LUP(LxL) = M^{-1}(LxL)
 *           LUP is alocated here, freed by caller.
 */
double *
Cal_LUP_Decomp(FILE *ofp, double *M, int *perm, int L, int verbose)
{
  double *LU;
  double *m;
  double  val;
  double  exchange;
  int     new_row, new_perm;
  int     L2;
  int     i, j, k;

  L2 = L * L;

  /* allocate LUP[L2]
   */
  LU = (double *) MallocOrDie (sizeof(double) * L2);
  m  = (double *) MallocOrDie (sizeof(double) * L2);

  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) {
      m[i*L+j]  = M[i*L+j];
      LU[i*L+j] = 0.;
    }

  for (i = 0; i < L; i++) 
    perm[i] = i; /* initializa to no permutations */

  /* start moving by columns
   */
  for (k = 0; k < L; k++) {

    val = 0.;
    /* For a given column move rows to have the largest possible value
     * in the diagonal.
     */
    for (i = k; i < L; i++) {
      if (fabs(m[i*L+k]) > val) {
	val = fabs(m[i*L+k]);
	new_row = i;
      }
    }
    if (val == 0.) Die( "this matrix is singular. No LUP decomposition possible");

    /* register in the P matrix that you did a permutation
     */
    new_perm      = perm[k];
    perm[k]       = perm[new_row];
    perm[new_row] = new_perm;
    
    /* exchange values of the two rows both for LU and m
     */
    for (i = 0; i < L; i++) {
      exchange       = m[k*L+i];
      m[k*L+i]       = m[new_row*L+i];
      m[new_row*L+i] = exchange;

      exchange        = LU[k*L+i];
      LU[k*L+i]       = LU[new_row*L+i];
      LU[new_row*L+i] = exchange;
    }
    
    LU[k*L+k] = m[k*L+k];
    
    for (i = k+1; i < L; i++) {
      LU[i*L+k] = m[i*L+k]/LU[k*L+k];  /* L part of the LUP matrix */
      LU[k*L+i] = m[k*L+i];            /* U part of the LUP matrix */
    }
    
    /* redefine rest of matrix as m_ij -= l_ik * u_kj
     */
    for (i = k+1; i < L; i++) 
      for (j = k+1; j < L; j++) 
	m[i*L+j] -= LU[i*L+k]*LU[k*L+j];
	
  }
  if (verbose) {
    fprintf(ofp, "LU matrix\n");
    PrintProbs(ofp, LU, L);

    fprintf(ofp, "L * U matrix\n");
    for (i = 0; i < L; i++) 
      for (j = 0; j < L; j++) 
	m[i*L+j] = 0.;

    for (i = 0; i < L; i++) 
      for (k = 0; k < L; k++) {
	if (i <= k) m[i*L+k] += LU[i*L+k];

	for (j = 0; j <= ((i-1 < k)?i-1:k); j++) 
	  m[i*L+k] += LU[i*L+j] * LU[j*L+k];
      }
    
    PrintProbs(ofp, m, L);
    
    fprintf(ofp, "P matrix\n");
    for (i = 0; i < L; i++) {
      for (j = 0; j < L; j++) 
	fprintf(ofp, "%.4f ", ((perm[i] == j)? 1. : 0.));
      fprintf(ofp, "\n");
    }
  }

  free(m);

  return LU;
}


/* Function: Cal_M_Exp()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  Given a matrix M, calculate exp{r*M}
 *
 *            exp{rM} = \sum_{n=0}^{\infty} [ r^n * M^n / n!   ] 
 *
 * Args:     M  - LL joint prob matrix (prealloc)
 *
 * Returns:  Q(LxL) = exp{rM(LxL)}
 *           Q is alocated here, freed by caller.
 */
double *
Cal_M_Exp(FILE *ofp, double *M, int L, double r, int verbose)
{
  double *Qr;        /* Qr = exp{r*M} matrix */
  double *taylorQr;  /* next term for Qr in the taylor expansion */
  double *Mpower;    /* holds at a given n (M_I)^n */
  double  coeff;
  int     L2;
  int     stop = FALSE;
  int     i;
  int     n = 0;
  
  L2 = L * L;

  /* allocate Qr[L2]
   */
  taylorQr = (double *) MallocOrDie (sizeof(double) * L2);

  /* inititalize Mpower to I and Qr to I
   */
  Mpower = Cal_Id(L);
  Qr     = Cal_Id(L);

  coeff  = 0.0; 
  while (!stop) {
    n++;

    coeff += log(r/n);
    
    /* calculate M^{n-1}*M
     */
    Comp_M_N_Prod(ofp, M, Mpower, L, FALSE); /* multipL Mpower*M */
    
    for (i = 0; i < L2; i++) {
      taylorQr[i] = exp(coeff) * Mpower[i];
      if (abs(taylorQr[i]) > DBL_MAX || abs(Qr[i]-taylorQr[i]) > DBL_MAX) 
	Die ("sorry I did not reach convergence before double-float limit in Cal_M_Exp()\n");
    }
    
    if (Check_Accuracy(taylorQr, L*L)) {
      for (i = 0; i < L2; i++) { 
	Qr[i] += taylorQr[i];
      }
    }
    else {
      stop = TRUE; 
      if (verbose) fprintf(ofp, "exp(r*K) convergence in %d terms\n", n);
    }   
  }
  
  if (verbose) {
    fprintf(ofp, "exp{r*M} matrix\n");
    PrintProbs(ofp, Qr, L);
  }

  free(Mpower);
  free(taylorQr);

  return Qr;
}

void
Comp_M_Exp(FILE *ofp, double *M, int L, double r, int verbose)
{
  double *Qr;        /* Qr = exp{r*M} matrix */

  Qr = Cal_M_Exp(ofp, M, L, r, verbose);

  CopyMatrix (M, Qr, L, L);

  if (verbose) {
    fprintf(ofp, "exp{r*M} matrix\n");
    PrintProbs(ofp, M, L);
  }

  free(Qr);
}

/* Function: Cal_M_Inv()
 * Date:     ER, Mon Apr 17 13:40:50 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculate M^{-1} using a LUP decomposition
 *
 * Method:   PA = LU => LUA^{-1} = P
 *
 * Args:     M^{-1} - LxL conditional prob matrix (prealloc)
 *
 * Returns:  Minv(LxL) = M^{-1}(LxL)
 *           Minv is alocated here, freed by caller.
 */
double *
Cal_M_Inv(FILE *ofp, double *M, int L, int verbose)
{
  double *Minv;     /* Minv  = M^{-1} matrix        */
  double *LU;       /* LU = L*U = PM matrix         */
  double *b;        /* LU = L*U = M matrix          */
  int    *perm;     /* P matrix                     */
  int     L2;
  int     i, j, k;
  double  sum;

  L2 = L * L;

  if (verbose ) {
    fprintf(ofp, "M matrix\n");
    PrintProbs(ofp, M, L);
 }

  /* allocate Minv[L2], LU
   */
  Minv  = (double *) MallocOrDie (sizeof(double) * L2);
  b     = (double *) MallocOrDie (sizeof(double) * L);
  perm  = (int    *) MallocOrDie (sizeof(int)    * L);

  LU = Cal_LUP_Decomp(ofp, M, perm, L, verbose);

  for (j = 0; j < L; j++) {
    for (i = 0; i < L; i++) 
      if (j == perm[i]) b[i] = 1.;
      else b[i] = 0.;
    
    Solve_LU_Eq(ofp, LU, b, L, verbose);

    for (i = 0; i < L; i++) 
      Minv[i*L+j] = b[i];
  }
  
  if (verbose ) {
     fprintf(ofp, "M^{-1} matrix\n");
     PrintProbs(ofp, Minv, L);

    fprintf(ofp, "M * M^{-1} matrix\n");
    for (i = 0; i < L; i++) {
      for (k = 0; k < L; k++) {
	sum = 0.;
	for (j = 0; j < L; j++) 
	  sum += M[i*L+j] * Minv[j*L+k];
	
	fprintf(ofp, "%.4f ", sum);
      }
      fprintf(ofp, "\n");
    }
  }
  
  free(LU);
  free(perm);
  free(b);

  return Minv;
}

void
Comp_M_Inv(FILE *ofp, double *M, int L, int verbose)
{
  double *Minv;     /* Minv  = M^{-1} matrix        */

  Minv = Cal_M_Inv(ofp, M, L, verbose);

  CopyMatrix (M, Minv, L, L);

  if (verbose) {
    fprintf(ofp, "M^{-1} matrix\n");
    PrintProbs(ofp, M, L);
  }

  free(Minv);
}

/* Function: Cal_M_Log()
 * Date:     ER, Wed Mar  1 11:04:24 CST 2000 [St. Louis]
 *
 * Purpose:  given M calculate K = log{M} using a taylor series.
 *
 *           K = \sum_{n=1}^{\infty} [ (-)^{n+1}(M-I)^n / n]
 *
 * Args:     Q  - LxL conditional prob matrix (prealloc)
 *
 * Returns:  K(LxL) = logM(LxL)
 *           K is alocated here, freed by caller.
 */
double *
Cal_M_Log(FILE *ofp, double *M, int L, int verbose)
{
  double *K;        /* K = logM matrix                         */
  double *taylorK;  /* next term for K in the taylor expansion */
  double *Mmin;     /* holds M - I                             */
  double *Mpower;   /* holds at a given n (M_I)^n              */
  double  coeff;
  int     L2;
  int     stop = FALSE;
  int     i;
  int     n = 0;

  L2 = L * L;

  /* allocate K[L2]
   */  
  K       = (double *) MallocOrDie (sizeof(double) * L2);
  taylorK = (double *) MallocOrDie (sizeof(double) * L2);

  /* substract I from M
   */
  Mmin = Cal_M_MinId(M, L);  /* M --> Mmin = M-I */

  if (verbose) {
    fprintf(ofp, "Mmin matrix\n");
    PrintProbs(ofp, Mmin, L);
  }

  /* inititalize Mpower to I and K to zero
   */
  Mpower = Cal_Id(L);
  for (i = 0; i < L2; i++) 
    K[i] = 0.0;
  
  while (!stop) {
    n ++;
    
    Comp_M_N_Prod(ofp, Mmin, Mpower, L, FALSE); /* multiply Mpower*Mmin */
    
    coeff = (n%2==1)? 1./n : -1./n;

    for (i = 0; i < L2; i++) {
      taylorK[i] = coeff * Mpower[i];
      if (taylorK[i] > DBL_MAX) 
	Die ("sorry I did not reach convergence before double-float limit in Cal_M_Log() \n");
    }
     
  if (FALSE) {
    fprintf(ofp, "logQ taylor matrix\n");
    PrintProbs(ofp, taylorK, L);
  }

    if (Check_Accuracy(taylorK, L*L)) {
      for (i = 0; i < L2; i++) 
	K[i] += taylorK[i];
    }
    else {
      stop = TRUE; 
      if (verbose) fprintf(ofp, "logM convergence in %d terms\n", n);
    }   
  }
  
  if (verbose) {
    fprintf(ofp, "logM matrix\n");
    PrintProbs(ofp, K, L);
  }
  
  free(taylorK);
  free(Mmin);
  free(Mpower);
  return K;
}

void
Comp_M_Log(FILE *ofp, double *M, int L, int verbose)
{
  double *K;        /* K = logM matrix                         */

  K = Cal_M_Log(ofp, M, L, verbose);

   CopyMatrix (M, K, L, L);

  if (verbose) {
    fprintf(ofp, "log{M} matrix\n");
    PrintProbs(ofp, M, L);
  }

 free(K);
}

/* Function: Cal_M_MinId()
 * Date:     ER, Thu Apr 13 10:30:38 CDT 2000 [St. Louis]
 *
 * Purpose:  given M calculates M - Id.
 *
 * Args:     M  - LxL (prealloc)
 *
 * Returns:  Mmin(LixLj) = M(LxL) - ID
 *           Mmin is alocated here, freed by caller.
 */
double *
Cal_M_MinId(double *M, int L)
{
  double *Mmin;
  int     i, j;

  Mmin = (double *) MallocOrDie (sizeof(double) * L * L);
  
  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      Mmin[i*L+j] =  M[i*L+j] - ((i==j)? 1. : 0);
  
  return Mmin;
}

void
Comp_M_MinId(double *M, int L)
{
  int i, j;

  for (i = 0; i < L; i++)
    for (j = 0; j < L; j++) 
      M[i*L+j] -= ((i==j)? 1. : 0);
  
}

/* Function: Cal_M_num_Prod()
 * Date:     ER,  [St. Louis]
 *
 * Purpose:  given M (LixLj) and N (LixLj) calculate Msum(LixLj) = M + N.
 *
 * Args:     M    - LixLj (prealloc)
 *           num  - a double
 *
 * Returns:  Mprod(LixLj) = M(LixLj) * num
 *           Mprod is alocated here, freed by caller.
 */
double *
Cal_M_num_Prod(FILE *ofp, double *M, double num, int Li, int Lj, int verbose)
{
  int     i, j;
  double *Mprod;
  
  Mprod = (double *) MallocOrDie (sizeof(double) * Li * Lj);
  
  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++)
	Mprod[i*Lj+j] = M[i*Lj+j] * num;
  
  if (verbose) {
    fprintf(ofp, "num * M matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", Mprod[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
  return Mprod;
}

void
Comp_M_num_Prod(FILE *ofp, double *M, double num, int Li, int Lj, int verbose)
{
  int  i, j;
  
  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++)
	M[i*Lj+j] *= num;
  
  if (verbose) {
    fprintf(ofp, "num * M matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", M[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
}


/* Function: Cal_M_N_Prod()
 * Date:     ER, Thu Apr 13 10:30:38 CDT 2000 [St. Louis]
 *
 * Purpose:  given M (LixLk) and N (LkxLj) calculate Mprod(LixLk) = M * N.
 *
 * Args:     M  - LixLk (prealloc)
 *           N  - LkxLj (prealloc)
 *
 * Returns:  Mprod(LixLj) = M(LixLk) * N(LkxLj)
 *           Mprod is alocated here, freed by caller.
 */
double *
Cal_M_N_Prod(FILE *ofp, double *M, double *N, int Li, int Lk, int Lj, int verbose)
{
  int     i, j, k;
  double  prod;
  double *Mprod;

  Mprod = (double *) MallocOrDie (sizeof(double) * Li * Lj);
  
  if (verbose) {
    fprintf(ofp, "M matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (k = 0; k < Lk; k++) 
	fprintf(ofp, "%.4f ", M[i*Lk+k]);
      fprintf(ofp, "\n");
    }

    fprintf(ofp, "N matrix\n");
    
    for (k = 0; k < Lk; k++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", N[k*Lj+j]);
      fprintf(ofp, "\n");
    }
  }

  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++) {
      prod = 0.;
      for (k = 0; k < Lk; k++) {
	prod += M[i*Lk+k] * N[k*Lj+j];
	if (prod > DBL_MAX) 
	  Die("sorry, I am getting out of bounds in Cal_M_N_Prod() prod[%d][%d] = %f\n", i, j, prod);
      }
      Mprod[i*Lj+j] = prod;
    }

  if (verbose) {
    fprintf(ofp, "M*N matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", Mprod[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
  return Mprod;
}

void
Comp_M_N_Prod(FILE *ofp, double *M, double *N, int L, int verbose)
{
  double *Mprod;

  Mprod = Cal_M_N_Prod(ofp, M, N, L, L, L, verbose);

  CopyMatrix (N, Mprod, L, L);

  if (verbose) {
    fprintf(ofp, "M*N matrix\n");
    PrintProbs(ofp, N, L);
  }

  free(Mprod);
}

/* Function: Cal_M_N_Sum()
 * Date:     ER, Mon May  8 10:44:33 CDT 2000 [St. Louis]
 *
 * Purpose:  given M (LixLj) and N (LixLj) calculate Msum(LixLj) = M + N.
 *
 * Args:     M  - LixLj (prealloc)
 *           N  - LixLj (prealloc)
 *
 * Returns:  Msum(LixLj) = M(LixLj) * N(LixLj)
 *           Msum is alocated here, freed by caller.
 */
double *
Cal_M_N_Sum(FILE *ofp, double *M, double *N, int Li, int Lj, int add, int verbose)
{
  int     i, j;
  double *Msum;
  
  Msum = (double *) MallocOrDie (sizeof(double) * Li * Lj);
  
  for (i = 0; i < Li; i++)
    for (j = 0; j < Lj; j++) 
      if (add)
	Msum[i*Lj+j] = M[i*Lj+j] + N[i*Lj+j];
      else
	Msum[i*Lj+j] = M[i*Lj+j] - N[i*Lj+j];
  
  if (verbose) {
    fprintf(ofp, "M+N matrix\n");
    
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", Msum[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }
  return Msum;
}

void
Comp_M_N_Sum(FILE *ofp, double *M, double *N, int Li, int Lj, int add, int verbose)
{
  int     i, j;
  double *Msum;
  
  Msum = Cal_M_N_Sum(ofp, M, N, Li, Lj, add, verbose);
  
  CopyMatrix (N, Msum, Li, Lj);

  if (verbose) {
    fprintf(ofp, "M + N matrix\n");
    for (i = 0; i < Li; i++) {
      for (j = 0; j < Lj; j++) 
	fprintf(ofp, "%.4f ", N[i*Lj+j]);
      fprintf(ofp, "\n");
    }
  }

  free(Msum);
}

double *
Col_From_M (double *A, int Lrow, int Lcol, int idx)
{
  double *col;
  int     row_i;

  if (idx >= Lcol) Die ("cannot give you that col");

  col = (double *) MallocOrDie (sizeof(double) * Lcol);
 
  for (row_i=0; row_i< Lrow; row_i++) col[row_i] = A[row_i*Lcol+idx];

  return col;
}

void
CopyMatrix (double *copyQ, double *Q, int N, int M)
{
  int col, row;

  for (row = 0; row < N; row++) 
    for (col = 0; col < M; col++) 
      copyQ[row*M+col] = Q[row*M+col];

}

int
IsDiagonal (double *A, int dim)
{
  int flag = 1;
  int row, col;
  
  for (row = 0; row < dim; row++) 
    for (col = 0; col < dim; col++) {
      if (col != row && (A[col*dim+row] > MARGIN || A[col*dim+row] < - MARGIN)) flag = 0; break;
    }

  return flag;
}

/* Function: HessenbergForm()
 *
 * Date:    ER, Fri Apr 21 16:15:23 CDT 2000 [St. Louis]
 *
 * Purpose: Given a real matrix M (LxL) obtain its Hessenberg form.
 *          This is an intermediate step to calculate the eigenvalues
 *          using the QL algorithm for real non-symmetric matrices.
 *
 *          The Hessenberg form has zeros everywhere below the diagonal,
 *          expect for the subdiagonal.
 *
 *
 *          Implemented after reading chapter 11 of "numerical recipies in C"
 *
 * Method:  -- pick a column k (k=0, k<L-1)
 *
 *          -- look at rows  i (i>k, i<L)
 *
 *                 find row i_o with larger value M(i_o,k)
 *
 *          -- exchange rows i_o <--> k+1
 *          -- exchange cols i_o <--> k+1
 *
 *          -- for rows  i (i=k+2, i<L)
 *
 *                 row_i <-- row_i - row_{k+1}*M(i,k)/M(k+1,k)
 *
 *             notice that for the new row_i:
 *                 M(i,k) = 0 so we have put zeros in column k, for all rows under the subdiagonal (i > k+1)
 *
 *         -- to make the elimination a similarity transformation, also reassign:
 *
 *               for rows  i (i=k+2, i<L)
 *                 col_{k+1} <-- col_{k+1} - col_{i}*M(i,k)/M(k+1,k)
 *
 * Args:              
 *
 * Returns:  H hessenberg form.
 *           H is allocated here, and freed by caller.
 *
 */
double *
HessenbergForm(FILE *ofp, double *M, int L)
{
  double *H;
  double  val;
  double  exchange;
  int     i, j, k, k1;
  int     new_row;


  H  = (double *) MallocOrDie (sizeof(double) * L * L);

  for (i = 0; i < L; i++) 
    for (j = 0; j < L; j++) 
      H[i*L+j] = M[i*L+j];

  /* start moving by columns
   */
  for (k = 0; k < L-1; k++) {

    k1 = k + 1;
    val = 0.;
    /* For a given column move rows to have the largest possible value
     * in the diagonal.
     */
    new_row = k1; /* initialize */
    for (i = k1; i < L; i++) {
      if (fabs(H[i*L+k]) > val) {
	val = fabs(H[i*L+k]);
	new_row = i;
      }
    }
    
    if (k1 != new_row) {
      for (i = 0; i < L; i++) {
	/* exchange values of the two rows (k+1 and new_row) 
	 */
	exchange       = H[k1*L+i];
	H[k1*L+i]      = H[new_row*L+i];
	H[new_row*L+i] = exchange;
      }
      for (i = 0; i < L; i++) {
	/* also exchange values for the columns (k+1 and new_row)
	 */
	exchange       = H[i*L+k1];
	H[i*L+k1]      = H[i*L+new_row];
	H[i*L+new_row] = exchange;
      }
    }
    
    if (val != 0.) {
      for (i = k1+1; i < L; i++) {
	for (j = 0; j < L; j++) 
	  H[i*L+j]  -= H[i*L+k] * H[k1*L+j] / val;
	for (j = 0; j < L; j++) 
	  H[j*L+k1] += H[i*L+k] * H[j*L+i] / val;
     }
    }
  } /* for every column k */
  
  return H;
}

/* Function: Hessenberg2Eigenvalues()
 *
 * Date:    ER, Fri Apr 21 21:07:27 CDT 2000 [St. Louis]
 *
 * Purpose: Given a real non-symmetric matrix M (LxL) in its Hessenberg form
 *          calculate its eigenvalues using the QR algorithm.
 *
 *          Implemented after reading chapter 6 of Matrix Methods (R. Bronson).
 *
 *          A_o = H
 *
 *          Use QR decomposition to calculate    A_o - (A_o)_nn I = Q_o * R_o
 *
 *          then,                                A_1              = R_o * Q_o + (A_o)_nn I
 *
 *          The QR decomposition preserves the Hessenberg form (which makes it faster).
 *
 *          At the end A_k is of the form       |   S      T |
 *                                              |            |
 *                                              | 0...0    a |    ===> a is an eigenvalue. Continue QR with submatrix S
 *
 *          OR
 *                                              |   S       T  |
 *                                              |              |
 *                                              | 0...0    b c |
 *                                              | 0...0    d e | ===> 2 complex eigenvalues. Continue QR with submatrix S
 *
 *
 * Args:    H -- (LxL) Hessenberg matrix          
 *
 * Returns: eigen(2L) vector with real and imaginary part of eigenvalues
 *
 */
void
Hessenberg2Eigenvalues(FILE *ofp, double *H, int L, struct eigenvalue_s **ret_eigen, int pedantic, int verbose)
{
  struct eigenvalue_s *eigen;
  double *A;         /* the series matrix that is equivalent to the original matrix */
  double *Q;
  double *R;
  double *I;
  double *last_row;
  double *nxtl_row;
  double  Ann;       /* nn element of the A matrix                                  */
  double  a, b, c;   /* coefficients of second degree equation ax^2+bx+c=0          */
  int     dim;       /* dimension of the square matrix A                            */
  int     it = 0;    /* number of iterations                                        */
  int     idx;       /* order of the eigenvalues being calculated                   */
  int     i;
  int     flag = 0;
  
  /* memory allocation */
  A =  (double *) MallocOrDie (sizeof(double) * L * L);
  Q =  (double *) MallocOrDie (sizeof(double) * L * L);
  R =  (double *) MallocOrDie (sizeof(double) * L * L);
  
  eigen       = (struct eigenvalue_s *) MallocOrDie (sizeof(struct eigenvalue_s));
  eigen->real = (double              *) MallocOrDie (sizeof(double) * L);
  eigen->imag = (double              *) MallocOrDie (sizeof(double) * L);
  
  
  /* initialize eigenvalues */
  for (i = 0; i < L; i++) {
    eigen->real[i] = 0.;
    eigen->imag[i] = 0.;
  }
  
  /* initialize A_o matrix Q and R*/
  for (i = 0; i < L*L; i++) {
    A[i] = H[i];
    Q[i] = 0.0;
    R[i] = 0.0;
  }
  
  /* initialize dimension of space */
  dim = L;
  
  /* initialize number of eigenvalues */
  idx = 0;
  
  /* do the iteration */
  
  while (dim > 2) {
    flag = 0;
    it ++;
    if (verbose) 
      printf("Hessenberg2Eigenvalues()\nnumber of iterations = %d\ndimension = %d\n", it, dim);
    
    last_row = Row_From_M (A, dim, dim, dim-1);
    nxtl_row = Row_From_M (A, dim, dim, dim-2);
    
    for (i=0; i<dim-1; i++) 
      if (last_row[i] > MARGIN || last_row[i] < -MARGIN) { flag = 1; break; }
    
    if (flag == 0) { /* one real eigenvalue */    
      eigen->real[idx] = last_row[dim-1];
      eigen->imag[idx] = 0.0;

      if (verbose) printf("  [%d]->real eigenvalue %f\n", idx, eigen->real[idx]);

      idx ++; /* one more eigenvalue */    
      dim --; /* reduce the dimension of the matrix by one */
    }
    else 
      {
	flag = 0;
	for (i=0; i<dim-2; i++) 
	  if (last_row[i] > MARGIN || last_row[i] < -MARGIN || 
	      nxtl_row[i] > MARGIN || nxtl_row[i] < -MARGIN  ) { flag = 1; break; }

	if (flag == 0) { /* two (posibly complex)  eigenvalues */   
	  a = 1.0;
	  b = - nxtl_row[dim-2] - last_row[dim-1];
	  c = nxtl_row[dim-2]*last_row[dim-1] - nxtl_row[dim-1]*last_row[dim-2];
	  
	  SecondDegree_Solutions (a, b, c, eigen, idx);
	  
	  if (verbose) {
	    printf("  [%d]->pair eigenvalue (%f, %f) \n", idx,   eigen->real[idx],   eigen->imag[idx]);
	    printf("  [%d]->pair eigenvalue (%f, %f) \n", idx+1, eigen->real[idx+1], eigen->imag[idx+1]);
	  }

	  idx += 2; /* two more eigenvalues */
	  dim -= 2; /* reduce the dimension of the matrix by 2 */

	}
	else { /* ok, do the actual QR decomposition */ 

	  /* shift matrix */
	  Ann = A[(dim-1)*dim+(dim-1)];
	  I   = Cal_Id (dim);
	  Comp_M_num_Prod (ofp, I, Ann, dim, dim, verbose);
	  
	  QR_Decomposition (ofp, A, Q, R, dim, pedantic, verbose);  /* QR decomposition of A      */
	  
	  Comp_M_N_Prod (ofp, R, Q, dim, verbose);               /* calculate new A            */
	  CopyMatrix(A, Q, dim, dim);
	  Comp_M_N_Sum (ofp, A, I, dim, dim, FALSE, verbose);    /* add the shift              */
	  CopyMatrix(A, I, dim, dim);
	}
      }
    
    free(last_row);
    free(nxtl_row);
    
  }

  if (dim == 2) {

    a = 1.0;
    b = - A[0] - A[3];
    c = A[0]*A[3] - A[1]*A[2];

    SecondDegree_Solutions (a, b, c, eigen, idx);

    if (verbose) {
      printf("  [%d]->pair eigenvalue (%f, %f) \n", idx,   eigen->real[idx],   eigen->imag[idx]);
      printf("  [%d]->pair eigenvalue (%f, %f) \n", idx+1, eigen->real[idx+1], eigen->imag[idx+1]);
    }
    idx += 2;  /* two eigenvalues */
  }
  else if (dim == 1) {
    eigen->real[idx] = A[0];
    eigen->imag[idx] = 0.0;

    if (verbose) printf("  [%d]->real eigenvalue %f\n", idx, eigen->real[idx]);
    idx ++; /* one eigenvalue */
  }

  /* paranoia */
  if (idx != L) Die ("You have not calculated all the eigenvalues\n");

  /* printout eigenvalues */
  if (verbose) 
    {
      fprintf(ofp, "Eigenvalues for Q\n"); 
      for (i = 0; i < L; i++)  
	fprintf(ofp, "%.4f %.4f \n", eigen->real[i], eigen->imag[i]);
    }
 
  free (A);
  free (Q);
  free (R);
  free (I);

  *ret_eigen = eigen;
}


/* Function: QR_Decomposition()
 *
 * Date:    ER, Thu Jul 25 14:35:13 CDT 2002  [St. Louis]
 *
 * Purpose: Given a real non-symmetric matrix M (nxn) 
 *          calculate the QR decomposition.
 *
 *          Implemented after reading chapter 6 of Matrix Methods (R. Bronson).
 *
 *          X = [x_1, ..., x_n]
 *
 *          for each i [1,...,n]
 *                          
 *              - for each j [i,...,n]
 *                          r_ij = <x_i,x_j>  ---->   r_i = (0,...,0,r_ii,...,r_in)
 *                          
 *              - q_i = 1/r_ii x_i
 *                          
 *              - for each j [i+1,...,n]
 *                         x_j = x_j - r_ij q_i
 *
 *         Then define Q = [q_n,...,q_n]
 *           
 *                         | r_1 |
 *                         |  .  |
 *                     R = |  .  |                      and X = QR
 *                         |  .  |
 *                         | r_n |
 *
 *         Then define new X = RQ
 *
 *
 * Args:    X -- (LxL) Hessenberg matrix     
 *          Q
 *          R
 *          L - dimension
 *
 * Returns: void. Calculates the new X matrix
 *
 */
void 
QR_Decomposition (FILE *ofp, double *X, double *Q, double *R, int n, int pedantic, int verbose)
{
  double *Check;
  double *Xdup;
  int     i, j, k;
  
  Xdup = (double *) MallocOrDie (sizeof(double) * n * n);

  for (i = 0; i < n*n; i++) Xdup[i] = X[i];
  
  for (i=0; i<n; i++)
    {
      /* 1. calculate Rii = sqrt <x_i,x_i>
       */
      for (k=0; k<n; k++) R[i*n+i] += X[k*n+i] * X[k*n+i];
      R[i*n+i] = sqrt(R[i*n+i]);
      
      /* 2. calculate q_i = 1/Rii x_i
       */
      for (k=0; k<n; k++)   Q[k*n+i] = X[k*n+i] / R[i*n+i];

      /* 3. calculate Rij = <x_j,q_i>
       */
      for (j=i+1; j<n; j++) 
	for (k=0; k<n; k++) R[i*n+j] += X[k*n+j] * Q[k*n+i];

      /* 4. redefinition vector x_j by x_j - Rij * q_i
       */
      for (j=i+1; j<n; j++) 
	for (k=0; k<n; k++) X[k*n+j] -= R[i*n+j] * Q[k*n+i];
    }
  
  /* check is X = QR ? */
  Check = Cal_M_N_Prod(ofp, Q, R, n, n, n, FALSE);
  
  if (verbose) {    
    fprintf(ofp, "QR decomposition\n");
    fprintf(ofp, "Q Matrix\n");
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(ofp, "%.4f ", Q[i*n+j]);
      }
      fprintf(ofp, "\n");
    }
    
    fprintf(ofp, "R Matrix\n");
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(ofp, "%.4f ", R[i*n+j]);
      }
      fprintf(ofp, "\n");
    }

    fprintf(ofp, "X Matrix\n");
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(ofp, "%.4f ", Xdup[i*n+j]);
      }
      fprintf(ofp, "\n");
    }

    fprintf(ofp, "\nQR Matrix\n");
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(ofp, "%.4f ", Check[i*n+j]);
      }
      fprintf(ofp, "\n");
    }
    fprintf(ofp, "\n");
    
  }
  
  /*
  for (i = 0; i < n*n; i++) 
    if (abs(Check[i]) > 0.0) {
      if (Check[i]/Xdup[i] < accuracy || Check[i]/Xdup[i] > 2.0-accuracy) Die ("Bad QR decomposition %d %f %f", i, Xdup[i], Check[i]);
    }
    else
      if (Check[i] - Xdup[i] > MARGIN || Check[i] - Xdup[i] < -MARGIN) Die ("Bad QR decomposition %d %f %f", i, Xdup[i], Check[i]);
  */

  free (Check);
  free (Xdup);
  
}

double *
Row_From_M (double *A, int Lrow, int Lcol, int idx)
{
  double *row;
  int     col_i;

  if (idx >= Lcol) Die ("cannot give you that row");

  row = (double *) MallocOrDie (sizeof(double) * Lrow);
 
  for (col_i=0; col_i< Lcol; col_i++) row[col_i] = A[idx*Lcol+col_i];

  return row;
}

  
  
/* Function: SecondDegree_Solutions()
 * Date:     ER, Thu Jul 25 13:22:05 CDT 2002 [St. Louis]
 *
 * Purpose:  calculate the 2 solutions of equation ax^2 + bx + c = 0
 *         
 * Args:     a
 *           b
 *           c
 *           eigen
 *
 * Returns: void
 *           
 */
void
SecondDegree_Solutions(double a, double b, double c, struct eigenvalue_s *eigen, int idx)
{
  double discriminant;
  double real;
  double imag;

  real = 0.0;
  imag = 0.0;
  discriminant = b*b - 4.0*a*c;

  if      (discriminant >= 0) real += sqrt(+discriminant);
  else if (discriminant <  0) imag += sqrt(-discriminant);

  b    /= 2.0 * a;
  real /= 2.0 * a;
  imag /= 2.0 * a;

  eigen->real[idx]   = -b + real;
  eigen->real[idx+1] = -b - real;

  eigen->imag[idx]   = +imag;
  eigen->imag[idx+1] = -imag;

}



/* Function: Solve_LUP_Eq()
 * Date:     ER, Mon Apr 17 17:22:31 CDT 2000 [St. Louis]
 *
 * Purpose:  Given LU x = b solve for x
 *  
 *           (Implemented after reading chapter 31 of "introduction to algorithms"
 *            by Cormen et al.)
 *
 * Method:   (1) solve equation  Ly = b for y
 *                     trivial because L is a lower triangular matrix.
 *
 *                     y_0 = b_0/l_00 = b_0
 *
 *                     y_i = 1/l_ii [ b_i - \sum_{j=1}^{i-1} l_ij*y_j ] for i=2,...,L-1
 *
 *                         = b_i - \sum_{j=1}^{i-1} l_ij*y_j 
 *
 *           (2) solve equation  Ux = y for x
 *                     trivial because U is an upper triangular matrix.
 *
 *                     x_{L-1} = y_{L-1}/u_{L-1,L-1}
 *
 *                     x_i = 1/u_ii [ y_i - \sum_{j=i+1}^{L-1} u_ij*x_j ] for i=L-2,...,0
 *
 *
 * Args:     LU - LxL LU  matrix.
 *
 * Returns:  (void)
 *           values for x are dumped in b.
 */
void
Solve_LU_Eq(FILE *ofp, double *LU, double *b, int L, int verbose)
{
  int     i, j;
  double  val, add;

  /* calculate y (dump values into b)
   */
  for (i = 1; i < L; i++) {
    val = b[i];
    add = 0;
      for (j = 0; j < i; j++)
	{
	  add -= LU[i*L+j] * b[j];
	}
    val += add;
    b[i] = val; 
  }

  /* calculate x (dump values into b)
   */
 b[L-1] /= LU[(L-1)*L+L-1];

  for (i = L-2; i >= 0; i--) {
    val = b[i];
    add = 0;
      for (j = i+1; j < L; j++)
	{
	  add -= LU[i*L+j] * b[j];
	}
    val += add;
    b[i] = val / LU[i*L+i];
  }

}

